#include <stdio.h>
#include <io.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <direct.h>

#include "stuff.h"

char buffer[0x50000];
int pos = 0;
int linepos = 0;

int File_Read(char const* name, cache* buffer)
{
	int length;
	byte *buf;
    FILE* f;

    f = fopen(name, "rb");

    if(!f)
        exit(1);

    fseek(f, 0, SEEK_SET);
    fseek(f, 0, SEEK_END);
    length = ftell(f);
    fseek(f, 0, SEEK_SET);

	buf = (byte*)malloc(length);

	fread(buf, length, 1, f);

	fclose(f);

	*buffer = buf;
	return length;
}

void File_Write(char const* name, char* source, int length)
{
    FILE* f;

    f = fopen(name, "w");
    
    if(!f)
    {
        printf("File_Write: Couldn't open file (%s)\n", name);
        exit(1);
    }
    
    fwrite(source, length, 1, f);

    pos = 0;
    memset(buffer, 0, 0x50000);
    linepos = 0;
}

void Buffer_AddString(char* str)
{
    int len = strlen(str);
    int spaces = 0;

    while(len--)
    {
        switch(*str)
        {
            case '\n':
                linepos = 0;
                buffer[pos++] = *str;
                break;
            case '\t':
                spaces = TABSPACING - (linepos % TABSPACING);
                while(spaces--)
                {
                    buffer[pos++] = ' ';
                    linepos++;
                }
                break;
            default:
                buffer[pos++] = *str;
                linepos++;
                break;
        }

        *str++;
    }
}

int qsort_actiontable(const int* a, const int* b) 
{
    if(*a == *b)
        return 0;
    else
        if(*a < *b)
            return -1;
        else
            return 1;
}

int main(int argc, char** argv)
{
    sprname_t   sprname[SPRITECOUNT];
    state_t     state[STATECOUNT];
    mobjinfo_t  mobj[MOBJCOUNT];
    cache       exedata;
    int         filesize;
    int         i;
    char        statename[STATECOUNT][32];
    int         sprinc[SPRITECOUNT];
    char**      mobjnames = classnames;
    char**      action = actionnames;
    int         action_hi;
    int*        actiontable;
    int         actiontablecount;
    int         atpos;

    if(!argv[1])
    {
        printf("no exe specified\n");
        return 0;
    }

    if(!argv[2] || !argv[3] || !argv[4])
    {
        printf("arg1 = sprite table offset\narg2 = mobjinfo table offset\narg3 = state tableoffset\n\n");
        return 0;
    }

    filesize = File_Read(argv[1], &exedata);
    memcpy(sprname, (exedata + atoi(argv[2])), sizeof(sprname_t) * SPRITECOUNT);
    memcpy(mobj, (exedata + atoi(argv[3])), sizeof(mobjinfo_t) * MOBJCOUNT);
    memcpy(state, (exedata + atoi(argv[4])), sizeof(state_t) * STATECOUNT);

    free(exedata);

    pos = 0;
    memset(buffer, 0, 0x50000);
    linepos = 0;
    memset(statename, 0, (sizeof(char)*STATECOUNT) * 32);
    memset(sprinc, 0, (sizeof(int)*SPRITECOUNT) * 8);

    // build state names
    sprintf(statename[0], "S_NULL");
    for(i = 1; i < STATECOUNT; i++)
    {
        char* name = sprname[state[i].sprite].sprname;
        int spridx = state[i].sprite;

        sprintf(statename[i], "S_%s_%02d", name, sprinc[spridx]++);
    }

    Buffer_AddString("typedef enum\n{\n");
    for(i = 0; i < SPRITECOUNT; i++)
    {
        char name[9];

        Buffer_AddString("\t");

        sprintf(name, "SPR_%s", sprname[i].sprname);
        Buffer_AddString(name);
        Buffer_AddString(",\n");
    }

    Buffer_AddString("\tNUMSPRITES\n\n");
    Buffer_AddString("} spritenum_t;\n\n");

    File_Write("sprinfo.h", buffer, pos);

    Buffer_AddString("char *sprnames[NUMSPRITES+1] =\n{");
    for(i = 0; i < SPRITECOUNT; i++)
    {
        if(!(i % 8))
            Buffer_AddString("\n\t");
        Buffer_AddString("\"");
        Buffer_AddString(sprname[i].sprname);
        Buffer_AddString("\", ");
    }

    Buffer_AddString("NULL\n};\n\n");

    File_Write("sprinfo.c", buffer, pos);

    Buffer_AddString("typedef enum\n{\n");
    for(i = 0; i < MOBJCOUNT; i++)
    {
        char name[16];

        Buffer_AddString("\t");
        Buffer_AddString(mobjinfonames[i][0]);
        Buffer_AddString(",\t\t");

        sprintf(name, "//%03d\n", i);
        Buffer_AddString(name);
    }

    Buffer_AddString("\tNUMMOBJTYPES\n\n");
    Buffer_AddString("} mobjtype_t;\n\n");

    File_Write("mobjinfo.h", buffer, pos);

    Buffer_AddString("mobjinfo_t mobjinfo[NUMMOBJTYPES] =\n{\n");
    for(i = 0; i < MOBJCOUNT; i++)
    {
        char tmp[128];

        sprintf(tmp, "\t{\t\t/*%s*/\n", mobjinfonames[i][0]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i,\t\t//doomednum\n", mobj[i].doomednum);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//spawnstate\n", statename[mobj[i].spawnstate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i,\t\t//spawnhealth\n", mobj[i].spawnhealth);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//seestate\n", statename[mobj[i].seestate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//seesound\n", sfxnames[mobj[i].seesound]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i,\t\t//reactiontime\n", mobj[i].reactiontime);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//attacksound\n", sfxnames[mobj[i].attacksound]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//painstate\n", statename[mobj[i].painstate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i,\t\t//painchance\n", mobj[i].painchance);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//painsound\n", sfxnames[mobj[i].painsound]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//meleestate\n", statename[mobj[i].meleestate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//missilestate\n", statename[mobj[i].missilestate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//crashstate\n", statename[mobj[i].crashstate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//deathstate\n", statename[mobj[i].deathstate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//xdeathstate\n", statename[mobj[i].xdeathstate]);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//deathsound\n", sfxnames[mobj[i].deathsound]);
        Buffer_AddString(tmp);
        if(mobj[i].speed > 65536)
            sprintf(tmp, "\t\t%i*FRACUNIT,\t\t//speed\n", mobj[i].speed>>16);
        else
            sprintf(tmp, "\t\t%i,\t\t//speed\n", mobj[i].speed);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i*FRACUNIT,\t\t//radius\n", mobj[i].radius>>16);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i*FRACUNIT,\t\t//height\n", mobj[i].height>>16);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i,\t\t//mass\n", mobj[i].mass);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%i,\t\t//damage\n", mobj[i].damage);
        Buffer_AddString(tmp);
        sprintf(tmp, "\t\t%s,\t\t//activesound\n", sfxnames[mobj[i].activesound]);
        Buffer_AddString(tmp);

        if(mobj[i].flags)
        {
            dword flags = mobj[i].flags;
            dword bit = 1;
            int j = 0;
            boolean first = true;

            Buffer_AddString("\t\t");

            do
            {
                if(flags & bit)
                {
                    dword mask = bit;
                    dword x = 0;

                    while(mask >>= 1)
                        x++;

                    if(!first)
                    {
                        if(!(j % 5))
                            Buffer_AddString("\n\t\t");

                        Buffer_AddString("|");
                    }

                    Buffer_AddString(mobjflagnames[x]);
                    j++;

                    if(first)
                        first = false;
                }

            } while(bit <<= 1 <= 0x80000000);

            Buffer_AddString(",\t\t//flags\n");
        }
        else
            Buffer_AddString("\t\t0,\t\t//flags\n");

        if(mobj[i].namepointer)
            sprintf(tmp, "\t\t\"%s\",\t\t//namepointer\n", *mobjnames++);
        else
            sprintf(tmp, "\t\tNULL,\t\t//namepointer\n");

        Buffer_AddString(tmp);
        Buffer_AddString("\t},\n\n");
    }

    Buffer_AddString("};\n\n");

    File_Write("mobjinfo.c", buffer, pos);

    Buffer_AddString("typedef enum\n{\n");
    for(i = 0; i < STATECOUNT; i++)
    {
        Buffer_AddString("\t");
        Buffer_AddString(statename[i]);
        Buffer_AddString(",\n");
    }

    Buffer_AddString("\tNUMSTATES\n\n");
    Buffer_AddString("} statenum_t;\n\n");

    File_Write("stateinfo.h", buffer, pos);

    action_hi = 0;
    for(i = 0; i < STATECOUNT; i++)
    {
        if(state[i].action > action_hi)
            action_hi = state[i].action;
    }

    Buffer_AddString("// Doesn't work with g++, needs actionf_p1\n");

    actiontablecount = 0;
    do
    {
        char name[64];
        
        if(*action == NULL)
            break;

        sprintf(name, "void %s();\n", *action);
        actiontablecount++;
        Buffer_AddString(name);

    } while(*action++);

    actiontable = (int*)malloc(sizeof(int)*actiontablecount);
    memset(actiontable, 0, sizeof(int)*actiontablecount);

    atpos = 0;
    for(i = 0; i < STATECOUNT; i++)
    {
        int j;
        boolean match = false;

        for(j = 0; j < actiontablecount; j++)
        {
            if(actiontable[j] == state[i].action)
            {
                match = true;
                break;
            }
        }

        if(!match)
            actiontable[atpos++] = state[i].action;
    }

    qsort(actiontable, actiontablecount, sizeof(int), qsort_actiontable);


    Buffer_AddString("\n\n#pragma warning(push)\n#pragma warning(disable:4113)\n\n");
    Buffer_AddString("state_t	states[NUMSTATES] =\n{\n\n");

    for(i = 0; i < STATECOUNT; i++)
    {
        char name[128];

        sprintf(name, "/*%s*/\t\t{ ", statename[i]);
        Buffer_AddString(name);

        sprintf(name, "SPR_%s, ", sprname[state[i].sprite].sprname);
        Buffer_AddString(name);
        
        sprintf(name, "%i, ", state[i].frame);
        Buffer_AddString(name);

        sprintf(name, "%i, ", state[i].tics);
        Buffer_AddString(name);

        if(state[i].action)
        {
            char** name2 = actionnames;
            int j;

            for(j = 0; j < actiontablecount; j++)
            {
                if(actiontable[j] == state[i].action)
                    break;

                *name2++;
            }
            sprintf(name, "{ %s }, ", *name2);
            Buffer_AddString(name);
        }
        else
            Buffer_AddString("{ NULL }, ");

        sprintf(name, "%s },\n", statename[state[i].nextstate]);
        Buffer_AddString(name);
    }

    Buffer_AddString("};\n\n");
    Buffer_AddString("#pragma warning(pop)\n\n\n");

    File_Write("stateinfo.c", buffer, pos);

    return 1;
}


